/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the License file in this	*/
/*	package.				*/
/*						*/
/*	Implement very sub level procedure to 	*/
/*	debug potential memory leak (or other)	*/
/*	trouble.				*/
/*						*/
/************************************************/
#include	<malloc.h>
#include	<stdlib.h>
#include	<stdio.h>
#include	<unistd.h>
#include	<time.h>
#include	<syslog.h>
#include	<execinfo.h>
#include	<sys/time.h>
#include	<memory.h>
#include	"dbgmem.h"

#undef	calloc
#undef	malloc
#undef	realloc
#undef	free
#undef	strdup


#define MXTB	4	/*max trace deep		*/

typedef	struct timeval	TIMEVAL;

typedef struct leaktyp LEAKTYP;
struct leaktyp	{	/*leak detection structure	*/
	TIMEVAL date;	/*allocation date		*/
	size_t size;	/*memory size			*/
	void *tb[MXTB]; /*back trace			*/
	void *ptr;	/*allocated memory		*/
	LEAKTYP *prev;	/*previous allocated memory	*/
	LEAKTYP *next;	/*next allocated memory		*/
	};

int memleak=DMFALSE;	/*flag to check for memory leak	*/

static LEAKTYP *memory;	/*allocated memory structure	*/
static long nummem;	/*number of leak structure	*/
static char cpath[300];	/*Application curpath		*/
/*

*/
/************************************************/
/*						*/
/*	Procedure to compare to "leaking" memory*/
/*	information, from the backtrace stand	*/
/*	point.					*/
/*						*/
/************************************************/
static int sametrace(const LEAKTYP *l1,const LEAKTYP *l2)

{
int inx;
int diff;

inx=0;
diff=0;
while ((inx<MXTB)&&(diff==0)) {
  diff=(int)(l1->tb[inx])-(int)(l2->tb[inx]);
  inx++;
  }
return diff;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to compare to "leaking" memory*/
/*	information				*/
/*						*/
/************************************************/
static int cmpleak(const void *p1,const void *p2)

{
int diff;
LEAKTYP *l1;
LEAKTYP *l2;

l1=((LEAKTYP **)p1)[0];
l2=((LEAKTYP **)p2)[0];
if ((diff=sametrace(l1,l2))==0)
  diff=l1->date.tv_sec-l2->date.tv_sec;
  if (diff==0)
    diff=l1->date.tv_usec-l2->date.tv_usec;
return diff;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to convert back trace address	*/
/*	to sources line information.		*/
/*						*/
/************************************************/
static void disline(LEAKTYP *leak,const char *path,FILE *fichier)

{
#define ADDR2LINE	"addr2line"

int i;
FILE *canal;
char command[1000];

(void) sprintf(command,"%s -e %s",ADDR2LINE,path);
for (i=0;i<MXTB;i++) {
  char str[20];

  if (leak->tb[i]==0)
    break;
  (void) sprintf(str," %08lx",(long)leak->tb[i]);
  (void) strcat(command,str);
  }
if ((canal=popen(command,"r"))!=(FILE *)0) {
  char line[200];

  while (fgets(line,sizeof(line)-1,canal)!=(char *)0) {
    (void) fprintf(fichier,"%s",line);
    }
  (void) fclose(canal);
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to assigne a Backtrace stack	*/
/*	(to trace back leak) to memory structure*/
/*						*/
/************************************************/
static void setbtrace(LEAKTYP *leak,int step)

{
#define MXSTEP	10

void *zone[MXSTEP];
int num;
int i,j;
struct timezone tz;

(void) memset(zone,'\000',sizeof(zone));
num=backtrace(zone,MXSTEP);
for (j=0,i=step+1;(j<MXTB)&&(i<num);i++,j++) {
  leak->tb[j]=zone[i];
  }
(void) gettimeofday(&(leak->date),&tz);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to report backtrace via syslog*/
/*						*/
/************************************************/
static void logbacktrace(int loglev,int step)

{
#define MXSTEP	10

void *zone[MXSTEP];
int num;
int i;

(void) memset(zone,'\000',sizeof(zone));
num=backtrace(zone,MXSTEP);
for (i=step;i<num;i++) {
  (void) syslog(loglev,"\tbt--> %08lx",(long)zone[i]);
  }
}
/*

*/
/************************************************/
/*						*/
/*	procedure to locate a 'leak memory'	*/
/*	among the list of memory.		*/
/*						*/
/************************************************/
static void *findmemory(void *ptr)

{
LEAKTYP *leak;
LEAKTYP *scan;

leak=(LEAKTYP *)0;
scan=memory;
while (scan!=(LEAKTYP *)0) {
  if (scan->ptr==ptr) {
    leak=scan;
    break;
    }
  scan=scan->next;
  }
return leak;
}
/*

*/
/************************************************/
/*						*/
/*	procedure to get a "leak proof" memory	*/
/*	email to a shell script			*/
/*						*/
/************************************************/
static void *getmemory(size_t size,int step)

{
LEAKTYP *leak;

leak=(LEAKTYP *)calloc(1,sizeof(LEAKTYP));
leak->size=size;
leak->ptr=malloc(size);
leak->next=memory;
(void) setbtrace(leak,step+1);
if (memory!=(LEAKTYP *)0)
  memory->prev=leak;
memory=leak;
nummem++;
return leak->ptr;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to memorise the current	*/
/*	application path, will be used to 	*/
/*	dump the memory stat.			*/
/*						*/
/************************************************/
void dbg_setcurpath(char *path)

{
(void) strncpy(cpath,path,sizeof(cpath));
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to assigne new memory area	*/
/*	while checking (if needed) for memory	*/
/*	leakage.				*/
/*						*/
/************************************************/
void *dbg_malloc(size_t size)

{
void *ptr;

if (memleak==DMTRUE) {
  ptr=getmemory(size,1);
  }
else
  ptr=malloc(size);
return ptr;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to assigne clean new memory 	*/
/*	while checking (if needed) for memory	*/
/*	leakage.				*/
/*						*/
/************************************************/
void *dbg_calloc(size_t nmemb,size_t size)

{
void *ptr;

if (memleak==DMTRUE) {
  ptr=getmemory(nmemb*size,1);
  (void) memset(ptr,'\000',nmemb*size);
  }
else
  ptr=calloc(nmemb,size);
return ptr;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to free memory from memry pool*/
/*	while checking (if needed) for memory	*/
/*	leakage.				*/
/*						*/
/************************************************/
void dbg_free(void *ptr)

{
if (memleak==DMTRUE) {
  LEAKTYP *leak;

  if ((leak=findmemory(ptr))!=(LEAKTYP *)0) {
    if (leak->prev!=(LEAKTYP *)0) {
      leak->prev->next=leak->next;
      }
    else {
      memory=leak->next;
      if (memory!=(LEAKTYP *)0)
        memory->prev=(LEAKTYP *)0;
      }
    if (leak->next!=(LEAKTYP *)0)
      leak->next->prev=leak->prev;
    (void) free(leak->ptr);
    (void) free(leak);
    nummem--;
    }
  else {
    char *strloc="subsys.c:dbg_free, Unable to find 'leak memory' for ptr '%08lx'=<%s>";

    (void) fprintf(stderr,strloc,(long)ptr,(char *)ptr);
    (void) syslog(LOG_DEBUG,strloc,(long)ptr,(char *)ptr);
    (void) logbacktrace(LOG_DEBUG,2);
    }
  }
else
  (void) free(ptr);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to reallocate memory from pole*/
/*	while checking (if needed) for memory	*/
/*	leakage.				*/
/*						*/
/************************************************/
void *dbg_realloc(void *ptr,size_t size)

{
if (memleak==DMTRUE) {
  if (ptr==(void *)0)
    ptr=getmemory(size,1);
  else {
    LEAKTYP *leak;

    if ((leak=findmemory(ptr))!=(LEAKTYP *)0) {
      leak->ptr=realloc(leak->ptr,size);
      ptr=leak->ptr;
      (void) setbtrace(leak,1);
      }
    else {
      char *strloc="subsys.c:dbg_realloc, Unable to find 'leak memory' for ptr '%08lx'=<%s>";

      (void) fprintf(stderr,strloc,(long)ptr,(char *)ptr);
      (void) syslog(LOG_DEBUG,strloc,(long)ptr,(char *)ptr);
      (void) logbacktrace(LOG_DEBUG,2);
      ptr=(void *)0;
      }
    }
  }
else
  ptr=realloc(ptr,size);
return ptr;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to allocate memory to		*/
/*	duplicate a string.			*/
/*						*/
/************************************************/
char *dbg_strdup(const char *str)

{
char *ptr;

if (memleak==DMTRUE) {
  ptr=getmemory(strlen(str)+3,1);
  (void) strcpy(ptr,str);
  }
else
  ptr=strdup(str);
return ptr;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to dump current memory	*/
/*	allocation and detect memory allocation	*/
/*						*/
/*	Data will be dumped on /tmp under	*/
/*	application name and process pid.	*/
/*						*/
/************************************************/
void dbg_dumpmem(const char *appname)

{
if (memleak==DMTRUE) {
  FILE *fichier;
  char filename[300];

  (void) sprintf(filename,"/tmp/%s.%05d",appname,getpid());
  if ((fichier=fopen(filename,"w"))==(FILE *)0)
    fichier=stderr;
  if (nummem>0) {
    LEAKTYP *leak;
    LEAKTYP **memsort;
    int i;
 
    memsort=(LEAKTYP **)calloc(nummem,sizeof(LEAKTYP *)); 
    i=0;
    leak=memory;
    while (leak!=(LEAKTYP *)0) {
      memsort[i]=leak;
      leak=leak->next;
      i++;
      }
    (void) fprintf(fichier,"%05ld memory allocation still open\n",nummem);
    (void) qsort(memsort,nummem,sizeof(LEAKTYP *),cmpleak);
    leak=(LEAKTYP *)0;
    for (i=0;i<nummem;i++) {
      #define STRTIME	"%Y-%m-%d %H:%M:%S"
      struct tm *tm;
      char strtime[40];

      if ((leak==(LEAKTYP *)0)||(sametrace(leak,memsort[i])!=0)) {
        (void) fprintf(fichier,"\n");
	leak=memsort[i];
	(void) disline(leak,cpath,fichier);
	}
      tm=localtime(&(memsort[i]->date.tv_sec));
      (void) strftime(strtime,sizeof(strtime),STRTIME,tm);
      (void) fprintf(fichier,"\t size=%05d age='%s.%06ld' ptr=%08lx->'%s'\n",
		     	     memsort[i]->size,strtime,memsort[i]->date.tv_usec,
			     (long)memsort[i]->ptr,(char *)memsort[i]->ptr);
      }
    }
  else {
#ifdef DEBUGMEM
    (void) fprintf(fichier,"No memory leak detected\n");
#else
    (void) fprintf(fichier,"No memory leak detector\n(comment in "
			   "'#define DEBUGMEM' within dbgmem.h)\n");
#endif
    }
  (void) fclose(fichier);
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to free memory allocated by	*/
/*	a library procedure (as scandir)	*/
/*	Memory allocated by such library routine*/
/*	are not known by leak detection.	*/
/*						*/
/************************************************/
void dbg_sysfree(void *ptr)

{
(void) free(ptr);
}
